# needed for recursive class definitions
from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional
from datetime import datetime
from starburst_data_products_client.shared.models import PaginatedJsonDataClass

@dataclass
class Role(PaginatedJsonDataClass):
    syncToken: Optional[str]
    userId: str
    email: str
    defaultRoleId: str
    createdOn: datetime
    scimManaged: bool
    directlyGrantedRoles: List[Role]
    allRoles: List[Role]



from dataclasses import dataclass
from typing import List
from starburst_data_products_client.shared.models import PaginatedJsonDataClass

@dataclass
class Cluster(PaginatedJsonDataClass):
    syncToken: str
    clusterId: str
    name: str
    cloudRegionId: str
    catalogRefs: List[str]
    idleStopMinutes: int
    batchCluster: bool
    warpSpeedCluster: bool
    minWorkers: int
    maxWorkers: int
    clusterState: str
    trinoUri: str


from dataclasses import dataclass
from starburst_data_products_client.shared.models import JsonDataClass


@dataclass
class UserData(JsonDataClass):
    isBookmarked: bool
    # TODO: check docs for additional fields



from dataclasses import dataclass

from starburst_data_products_client.sep.data import View
from starburst_data_products_client.sep.data import MaterializedView
from starburst_data_products_client.sep.data import Owner
from starburst_data_products_client.sep.data import AccessMetadata
from starburst_data_products_client.sep.data import UserData
from starburst_data_products_client.sep.data import RelevantLinks

from typing import List, Optional
from datetime import datetime
from starburst_data_products_client.shared.models import JsonDataClass


@dataclass
class DataProduct(JsonDataClass):
    id: Optional[str]
    name: str
    catalogName: str
    schemaName: str
    dataDomainId: str
    summary: str
    description: Optional[str]
    createdBy: Optional[str]
    status: Optional[str]
    views: List[View]
    materializedViews: List[MaterializedView]
    owners: List[Owner]
    productOwners: Optional[List[Owner]]
    relevantLinks: List[RelevantLinks]
    createdAt: Optional[datetime]
    updatedAt: Optional[datetime]
    updatedBy: Optional[str]
    publishedAt: Optional[datetime]
    publishedBy: Optional[str]
    accessMetadata: Optional[AccessMetadata]
    ratingsCount: Optional[int]
    userData: Optional[UserData]
    matchesTrinoDefinition: Optional[bool]
    bookmarkCount: Optional[int]

@dataclass
class DataProductParameters(JsonDataClass):
    name: str
    catalogName: str
    schemaName: str
    dataDomainId: str
    summary: str
    description: Optional[str]
    views: List[View]
    materializedViews: List[MaterializedView]
    owners: List[Owner]
    relevantLinks: List[RelevantLinks]
    


from dataclasses import dataclass
from typing import List
from datetime import datetime
from starburst_data_products_client.shared.models import PaginatedJsonDataClass


@dataclass
class UserPrincipal(PaginatedJsonDataClass):
    id: str
    type: str


@dataclass
class UserRole(PaginatedJsonDataClass):
    roleName: str
    roleId: str
    principal: UserPrincipal
    adminOption: bool


@dataclass
class User(PaginatedJsonDataClass):
    syncToken: str
    userId: str
    email: str
    defaultRoleId: str
    createdOn: datetime
    scimManaged: bool
    directlyGrantedRoles: List[UserRole]
    allRoles: List[UserRole]



from dataclasses import dataclass
from typing import List
from starburst_data_products_client.galaxy.models import Catalog
from starburst_data_products_client.shared.models import PaginatedJsonDataClass


@dataclass
class CatalogSearchResult(PaginatedJsonDataClass):
    clusters: List[Catalog]



from dataclasses import dataclass
from datetime import datetime
from typing import Optional
from starburst_data_products_client.shared.models import JsonDataClass


@dataclass
class AccessMetadata(JsonDataClass):
    lastQueriedAt: Optional[datetime]
    lastQueriedBy: Optional[str]



from dataclasses import dataclass
from typing import List, Optional
from starburst_data_products_client.galaxy.models import Contact, Link
from starburst_data_products_client.shared.models import PaginatedJsonDataClass

@dataclass
class TagIdentifier(PaginatedJsonDataClass):
    tagId: str
    name: str


@dataclass
class RoleIdentifier(PaginatedJsonDataClass):
    roleId: str
    roleName: str


@dataclass
class SchemaMetadata(PaginatedJsonDataClass):
    schemaId: str
    description: Optional[str]
    owner: RoleIdentifier
    tags: List[TagIdentifier]
    contacts: List[Contact]
    links: List[Link]



import requests
from typing import List, cast
from starburst_data_products_client.galaxy.models import (CreateDataProductRequest, Cluster, User, Contact, Catalog, DataProduct, Tag,
                                        CreateDataProductResponse, TagIdentifier, TagResponse, SchemaMetadata)
from urllib.parse import quote_plus
from starburst_data_products_client.shared.api import ApiException


class Api:
    def __init__(self,
                 client_id: str,
                 client_secret: str,
                 username: str, password: str,
                 host: str,
                 cluster: str,
                 default_catalog: str,
                 api_key: str = None):
        self.client_id = client_id
        self.client_secret = client_secret
        self.username = username
        self.password = password
        self.url = host
        self.cluster = cluster
        self.default_catalog = default_catalog
        if api_key is not None:
            self.api_key = api_key
        else:
            self.api_key = self.__get_api_key__(url=host, client_id=client_id, client_secret=client_secret)

        self.__data_products_list__ = None
        self.__default_catalog_id__: str = None
        self.__users_list__ = None

    def get_default_catalog_id(self) -> str:
        if self.__default_catalog_id__ is None:
            self.__default_catalog_id__ = self.get_catalog_id_for_name(self.default_catalog)
        return self.__default_catalog_id__

    def __get_api_key__(self, url: str, client_id: str, client_secret: str) -> str:
        api_key_response = requests.post(url=f'{url}/oauth/v2/token',
                                         auth=(client_id, client_secret),
                                         headers={'Content-Type': 'application/x-www-form-urlencoded'},
                                         data='grant_type=client_credentials')
        return api_key_response.json()['access_token']

    def delete_data_product(self, data_product_name: str = None, data_product_id: str = None):
        if data_product_id is None:
            if data_product_name is None:
                raise ValueError('Either data product name or id must be provided')
            else:
                data_product_id = self.get_data_product_id_by_name(data_product_name)
        PATH = f'public/api/v1/dataProduct/{data_product_id}'
        response = requests.delete(
            f'{self.url}/{PATH}',
            headers={'Authorization': f'bearer {self.api_key}'}
        )

        if not response.ok:
            raise ApiException(
                f'Could not delete data product: {response.reason}. Status code {response.status_code}',
                reason=response.reason,
                status=response.status_code,
                body=response.text
            )

    def create_data_product(self, data_product_request: CreateDataProductRequest):
        PATH = 'public/api/v1/dataProduct'
        response = requests.post(
            f'{self.url}/{PATH}',
            headers={'Authorization': f'bearer {self.api_key}', "Content-type": "application/json"},
            data=data_product_request.to_json()
        )
        if not response.ok:
            raise ApiException(
                f'Could not create data product: {response.reason}. Status code {response.status_code}',
                reason=response.reason,
                status=response.status_code,
                body=response.text
            )
        create_data_product_response = CreateDataProductResponse.load(response.json())
        return create_data_product_response.dataProductId

    def update_data_product(self, data_product_request: CreateDataProductRequest):
        data_product_id = self.get_data_product_id_by_name(data_product_request.name)
        PATH = f'public/api/v1/dataProduct/{data_product_id}'
        response = requests.patch(
            f'{self.url}/{PATH}',
            headers={'Authorization': f'bearer {self.api_key}', "Content-type": "application/json"},
            data=data_product_request.to_json()
        )
        if not response.ok:
            raise ApiException(
                f'Could not create data product: {response.reason}. Status code {response.status_code}',
                reason=response.reason,
                status=response.status_code,
                body=response.text
            )

    def get_data_product_id_by_name(self, data_product_name: str) -> str:
        return next(iter([dp.dataProductId for dp in self.list_data_products() if dp.name == data_product_name]))

    def list_data_products(self, tag_name=None) -> List[DataProduct]:
        PATH = 'public/api/v1/dataProduct'
        if self.__data_products_list__ is None:
            self.__data_products_list__ = DataProduct.paginated_response_to_list(
                f'{self.url}/{PATH}',
                headers={'Authorization': f'bearer {self.api_key}'})

        if tag_name is not None:
            tagged_schema_names = [schema.schemaId for schema in
                                   self.get_tagged_schemas(self.get_default_catalog_id(), tag_name)]
            return [data_product for data_product in self.__data_products_list__
                    if data_product.schemaName in tagged_schema_names]
        return self.__data_products_list__

    def get_schema_tags(self, catalog_id: str, schema_name: str) -> List[TagIdentifier]:
        PATH = f'public/api/v1/catalog/{catalog_id}/schema'
        schema_metadata_response = SchemaMetadata.paginated_response_to_list(
            f'{self.url}/{PATH}',
            headers={'Authorization': f'bearer {self.api_key}'})
        return next(iter([schema.tags for schema in schema_metadata_response if schema.schemaId == schema_name]))

    def get_tagged_schemas(self, catalog_id: str, tag_name: str) -> List[SchemaMetadata]:
        PATH = f'public/api/v1/catalog/{catalog_id}/schema'
        schema_metadata_response = SchemaMetadata.paginated_response_to_list(
            f'{self.url}/{PATH}',
            headers={'Authorization': f'bearer {self.api_key}'})
        return [schema for schema in schema_metadata_response if tag_name in [tag.name for tag in schema.tags]]

    def __list_catalogs__(self) -> List[Catalog]:
        PATH = 'public/api/v1/catalog'
        return Catalog.paginated_response_to_list(
            f'{self.url}/{PATH}',
            headers={'Authorization': f'bearer {self.api_key}'})

    def get_catalog_id_for_name(self, name: str) -> str:
        catalogs = self.__list_catalogs__()
        return next(iter([catalog.catalogId for catalog in catalogs if catalog.catalogName == name]), None)

    def __list_clusters__(self) -> List[Cluster]:
        PATH = 'public/api/v1/cluster'
        return Cluster.paginated_response_to_list(
            f'{self.url}/{PATH}',
            headers={'Authorization': f'bearer {self.api_key}'})

    def get_cluster_by_name(self, name: str) -> Cluster:
        encoded_name = f'name={quote_plus(name)}'
        PATH = f'public/api/v1/cluster/{encoded_name}'
        cluster = Cluster.load(requests.get(f'{self.url}/{PATH}',
                                            headers={'Authorization': f'bearer {self.api_key}'}
                                            ).json())
        return cluster

    def create_schema_tag_str(self, tag_name: str):
        self.create_schema_tag(Tag(name=tag_name, color='green', description='Added by dp sync'))

    def create_schema_tag(self, tag: Tag):
        PATH = 'public/api/v1/tag'
        requests.post(f'{self.url}/{PATH}',
                      headers={'Authorization': f'bearer {self.api_key}'},
                      data=tag.to_json())

    def get_tag_by_name(self, tag_name: str) -> Tag:
        encoded_tag_name = f'name={quote_plus(tag_name)}'
        PATH = f'public/api/v1/tag/{encoded_tag_name}'
        return cast(Tag, TagResponse.load(requests.get(f'{self.url}/{PATH}',
                                                       headers={
                                                                'Authorization': f'bearer {self.api_key}'}).json()))

    def tag_schema(self, tag_id: str, catalog_id: str, schema_name: str):
        PATH = f'public/api/v1/tag/{tag_id}/catalog/{catalog_id}/schema/{quote_plus(schema_name)}'
        response = requests.put(f'{self.url}/{PATH}',
                                headers={'Authorization': f'bearer {self.api_key}'})

    def __list_users__(self) -> List[User]:
        PATH = 'public/api/v1/user'
        if self.__users_list__ is None:
            self.__users_list__ = User.paginated_response_to_list(
                f'{self.url}/{PATH}',
                headers={'Authorization': f'bearer {self.api_key}'})
        return self.__users_list__

    def emails_to_users(self, emails: List[str]) -> List[Contact]:
        return [Contact(userId=user.userId, email=user.email)
                for user in self.__list_users__() if user.email in emails]

    def __get_user_by_email(self, email: str) -> User:
        encoded_name = f'email={quote_plus(email)}'

        PATH = f'public/api/v1/user/{encoded_name}'
        return User.load(requests.get(f'{self.url}/{PATH}',
                                                 headers={'Authorization': f'bearer {self.api_key}'}
                                                 ).json())



class ApiException(Exception):
    def __init__(self, message, reason: str, status: int, body: str):
        super().__init__(message)
        self.reason = reason
        self.status = status
        self.body = body



from dataclasses import dataclass
from datetime import datetime
from typing import Optional
from starburst_data_products_client.shared.models import JsonDataClass


@dataclass
class DataProductSearchResult(JsonDataClass):
    id: str
    name: str
    catalogName: str
    schemaName: str
    dataDomainId: str
    summary: str
    description: Optional[str]
    createdBy: str
    status: str
    createdAt: datetime
    updatedAt: datetime
    publishedAt: Optional[datetime]
    publishedBy: Optional[str]
    lastQueriedAt: Optional[datetime]
    lastQueriedBy: Optional[str]
    ratingsCount: int
    userData: dict



from typing import cast, Self, List, Any
import marshmallow_dataclass
from dataclasses import asdict, is_dataclass
from json import JSONEncoder, dumps
from datetime import datetime
from functools import singledispatchmethod
from dataclasses import dataclass
from requests import get


class DateTimeEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat()
        return super().default(o)


class JsonDataClass:

    @singledispatchmethod
    @classmethod
    def load(cls, arg):
        raise NotImplemented('JsonDataClass.load() only accepts dicts and json strings')

    @load.register
    @classmethod
    def _(cls, data: str) -> Self:
        if is_dataclass(cls):
            schema = marshmallow_dataclass.class_schema(cls)()
            return cast(cls, schema.loads(data))
        else:
            raise NotImplemented('Only dataclasses should inherit from JsonDataClass!')

    @load.register
    @classmethod
    def _(cls, data: dict) -> Self:
        if is_dataclass(cls):
            schema = marshmallow_dataclass.class_schema(cls)()
            return cast(cls, schema.load(data))
        else:
            raise NotImplemented('Only dataclasses should inherit from JsonDataClass!')

    def asdict(self) -> dict:
        if is_dataclass(self):
            return asdict(self)
        else:
            raise NotImplemented('Only dataclasses should inherit from JsonDataClass!')

    def to_json(self) -> str:
        return dumps(self.asdict(), cls=DateTimeEncoder)


@dataclass
class PaginatedResponse(JsonDataClass):
    nextPageToken: str
    result: List[Any]


class PaginatedJsonDataClass(JsonDataClass):
    @classmethod
    def paginated_response_to_list(cls, url: str, headers: {}) -> List[Self]:
        response = get(url=url, headers=headers)
        paginated_response = PaginatedResponse.load(response.json())
        results = [cls.load(r) for r in paginated_response.result]
        while paginated_response.nextPageToken is not None and paginated_response.nextPageToken.strip() != '':
            response = get(url=url, headers=headers, params={'pageToken': paginated_response.nextPageToken})
            paginated_response = PaginatedResponse.load(response.json())
            results += [cls.load(r) for r in paginated_response.result]
        return results



from dataclasses import dataclass
from starburst_data_products_client.shared.models import PaginatedJsonDataClass

@dataclass
class Catalog(PaginatedJsonDataClass):
    catalogId: str
    catalogName: str



from dataclasses import dataclass
from typing import Optional
from starburst_data_products_client.shared.models import PaginatedJsonDataClass


@dataclass
class TagResponse(PaginatedJsonDataClass):
    syncToken: str
    tagId: str
    name: str
    color: str
    description: Optional[str]



from .Catalog import Catalog
from .ClusterSearchResult import CatalogSearchResult
from .Cluster import Cluster
from .Contact import Contact
from .Link import Link
from .Role import Role
from .User import User
from .Tag import Tag
from .TagResponse import TagResponse
from .SchemaMetadata import SchemaMetadata, TagIdentifier, RoleIdentifier
from .CreateDataProductRequest import CreateDataProductRequest
from .DataProduct import DataProduct
from .CreateDataProductResponse import CreateDataProductResponse



from dataclasses import dataclass
from starburst_data_products_client.sep.data import Column
from starburst_data_products_client.sep.data import MaterializedViewProperties

from typing import List, Optional
from datetime import datetime
from starburst_data_products_client.shared.models import JsonDataClass


@dataclass
class MaterializedView(JsonDataClass):
    name: str
    description: Optional[str]
    createdBy: Optional[str]
    definitionQuery: str
    definitionProperties: Optional[MaterializedViewProperties]
    status: Optional[str]
    columns: Optional[List[Column]]
    markedForDeletion: Optional[bool]
    createdAt: Optional[datetime]
    updatedAt: Optional[datetime]
    updatedBy: Optional[str]
    publishedAt: Optional[datetime]
    publishedBy: Optional[str]
    matchesTrinoDefinition: Optional[bool]



from dataclasses import dataclass
from typing import List
from starburst_data_products_client.galaxy.models import Contact, Link
from starburst_data_products_client.shared.models import PaginatedJsonDataClass


@dataclass
class CreateDataProductRequest(PaginatedJsonDataClass):
    name: str
    summary: str
    description: str
    catalogId: str
    schemaName: str
    contacts: List[Contact]
    links: List[Link]
    defaultClusterId: str



from dataclasses import dataclass

from starburst_data_products_client.sep.data import View
from starburst_data_products_client.sep.data import MaterializedView
from starburst_data_products_client.sep.data import Owner
from starburst_data_products_client.sep.data import AccessMetadata
from starburst_data_products_client.sep.data import UserData
from starburst_data_products_client.sep.data import RelevantLinks

from typing import List, Optional, Any
from datetime import datetime
from starburst_data_products_client.shared.models import JsonDataClass


@dataclass
class Domain(JsonDataClass):
    id: str
    name: str
    description: Optional[str]
    schemaLocation: Optional[str]
    assignedDataProducts: List[Any]
    createdBy: str
    createdAt: datetime
    updatedAt: datetime
    updatedBy: str



from dataclasses import dataclass
from typing import List, Optional
from datetime import datetime
from starburst_data_products_client.shared.models import JsonDataClass


@dataclass
class MaterializedViewImportMetadata(JsonDataClass):
    status: Optional[str]
    scheduledTime: Optional[datetime]
    startTime: Optional[datetime]
    finishTime: Optional[datetime]
    rowCount: Optional[int]
    error: Optional[str]



#  a class must be imported before any classes that references it
from .MaterializedViewProperties import MaterializedViewProperties
from .DataProductSearchResult import DataProductSearchResult
from .UserData import UserData
from .Column import Column
from .MaterializedView import MaterializedView
from .Owner import Owner
from .View import View
from .AccessMetadata import AccessMetadata
from .RelevantLinks import RelevantLinks
from .DataProduct import DataProduct, DataProductParameters
from .DataProductWorkflowStatus import DataProductWorkflowStatus
from .Domain import Domain
from .Tag import Tag
from .SampleQuery import SampleQuery
from .MaterializedViewImportMetadata import MaterializedViewImportMetadata
from .MaterializedViewRefreshMetadata import MaterializedViewRefreshMetadata



from dataclasses import dataclass
from starburst_data_products_client.sep.data import MaterializedViewImportMetadata
from typing import List, Optional
from datetime import datetime
from starburst_data_products_client.shared.models import JsonDataClass


@dataclass
class MaterializedViewRefreshMetadata(JsonDataClass):
    lastImport: Optional[MaterializedViewImportMetadata]
    incrementalColumn: Optional[str]
    refreshInterval: Optional[str]
    storageSchema: Optional[str]
    estimatedNextRefreshTime: Optional[datetime]



from dataclasses import dataclass
from typing import Optional
from starburst_data_products_client.shared.models import JsonDataClass


@dataclass
class SampleQuery(JsonDataClass):
    name: str
    description: Optional[str]
    query: str



from dataclasses import dataclass
from typing import Optional
from starburst_data_products_client.shared.models import JsonDataClass


@dataclass
class MaterializedViewProperties(JsonDataClass):
    refresh_interval: Optional[str]
    grace_period: Optional[str]



from starburst_data_products_client.sep.data import DataProductSearchResult
from starburst_data_products_client.sep.data import DataProduct, DataProductParameters
from starburst_data_products_client.sep.data import DataProductWorkflowStatus
from starburst_data_products_client.sep.data import Domain
from starburst_data_products_client.sep.data import MaterializedViewRefreshMetadata
from starburst_data_products_client.sep.data import SampleQuery
from starburst_data_products_client.sep.data import Tag

import requests
from typing import List
from json import dumps


class Api:
    """A client for interacting with the Starburst Enterprise Data Products API.

    This class provides methods to interact with various endpoints of the Starburst Data Products API,
    including data product management, domain management, tag management, and workflow operations.

    Attributes:
        DOMAIN_PATH (str): API endpoint path for domain operations
        DATA_PRODUCT_PATH (str): API endpoint path for data product operations
        DATA_PRODUCT_TAGS_PATH (str): API endpoint path for tag operations
    """

    DOMAIN_PATH = 'api/v1/dataProduct/domains'
    DATA_PRODUCT_PATH = 'api/v1/dataProduct/products'
    DATA_PRODUCT_TAGS_PATH = 'api/v1/dataProduct/tags'

    def __init__(self, host: str, username: str, password: str, protocol: str = 'https'):
        """Initialize the API client.

        Args:
            host (str): The hostname of the Starburst Data Products server
            username (str): Username for authentication
            password (str): Password for authentication
            protocol (str, optional): The protocol to use (http/https). Defaults to 'https'.

        Raises:
            ValueError: If the hostname includes a protocol
        """
        if '://' in host:
            raise ValueError(f'Hostname should not include protocol')
        self.host = host
        self.username = username
        self.password = password
        self.protocol = protocol


    # --- data product API methods ---
    def search_data_products(self, search_string: str=None) -> List[DataProductSearchResult]:
        """Search for data products matching the given search string.

        The search is performed against all data product attributes and is case-insensitive.
        The search string is bookended by '%' for partial matching.

        Args:
            search_string (str, optional): The string to search for. Defaults to None.

        Returns:
            List[DataProductSearchResult]: List of matching data product search results

        Raises:
            Exception: If the API request fails
        """
        #REQUEST searchOptions.searchString is bookended by '%' and compared against all
        #dp attributes https://github.com/starburstdata/starburst-enterprise/blob/807dbbbfb48b7e5ea87777fc3aae8cd360dea1e8/core/starburst-server-main/src/main/java/com/starburstdata/presto/dataproduct/search/SearchSqlBuilder.java#L213

        params = {'searchOptions': dumps({'searchString': search_string})} if search_string is not None else None
        response = requests.get(
            url=f'{self.protocol}://{self.host}/{self.DATA_PRODUCT_PATH}',
            auth=(self.username, self.password),
            params=params
        )
        if not response.ok:
            raise Exception(f'Request returned code {response.status_code}.\nResponse body: {response.text}')

        return [search_result for search_result in
                [DataProductSearchResult.load(result) for result in response.json()]
                if search_string is None or search_string in search_result.name]
    
    
    def create_data_product(self, data_product: DataProductParameters) -> DataProduct:
        """Create a new data product.

        Args:
            data_product (DataProductParameters): The parameters for the new data product

        Returns:
            DataProduct: The created data product

        Raises:
            Exception: If the API request fails
        """
        response = requests.post(
            url=f'{self.protocol}://{self.host}/{self.DATA_PRODUCT_PATH}',
            auth=(self.username, self.password),
            json=data_product.asdict()
        )
        if not response.ok:
            raise Exception(f'Request returned code {response.status_code}.\nResponse body: {response.text}')
        return DataProduct.load(response.json())

    
    def clone_data_product(
        self,
        dp_id: str,
        catalog_name: str,
        new_schema_name: str,
        new_name: str,
        domain_id: str=None
    ) -> DataProduct:
        """Clone an existing data product.

        Args:
            dp_id (str): ID of the data product to clone
            catalog_name (str): Name of the catalog for the cloned product
            new_schema_name (str): Name of the schema for the cloned product
            new_name (str): Name for the cloned product
            domain_id (str, optional): ID of the domain for the cloned product. Defaults to None.

        Returns:
            DataProduct: The cloned data product

        Raises:
            Exception: If the API request fails
        """
        body={
            'catalogName': catalog_name,
            'newSchemaName': new_schema_name,
            'newName': new_name
        }
        if domain_id is not None:
            body['dataDomainId'] = domain_id
        response = requests.post(
            url=f'{self.protocol}://{self.host}/{self.DATA_PRODUCT_PATH}/{dp_id}/clone',
            auth=(self.username, self.password),
            json=body
        )
        if not response.ok:
            raise Exception(f'Request returned code {response.status_code}.\nResponse body: {response.text}')
        return DataProduct.load(response.json())
    

    def get_data_product(self, dp_id: str) -> DataProduct:
        """Get details of a specific data product.

        Args:
            dp_id (str): ID of the data product to retrieve

        Returns:
            DataProduct: The requested data product

        Raises:
            Exception: If the API request fails
        """
        response = requests.get(
            url= f'{self.protocol}://{self.host}/{self.DATA_PRODUCT_PATH}/{dp_id}',
            auth=(self.username, self.password)
        )
        if not response.ok:
            raise Exception('bad request' + str(response))
        return DataProduct.load(response.json())
    
    
    def update_sample_queries(self, dp_id: str, sample_queries: List[SampleQuery]):
        """Update the sample queries for a data product.

        Args:
            dp_id (str): ID of the data product
            sample_queries (List[SampleQuery]): List of sample queries to set

        Raises:
            Exception: If the API request fails
        """
        response = requests.put(
            url=f'{self.protocol}://{self.host}/{self.DATA_PRODUCT_PATH}/{dp_id}/sampleQueries',
            auth=(self.username, self.password),
            json=[{'name':query.name,'description':query.description,'query':query.query} for query in sample_queries]
        )
        if not response.ok:
            raise Exception(f'Request returned code {response.status_code}.\nResponse body: {response.text}')
    
    
    def list_sample_queries(self, dp_id: str) -> List[SampleQuery]:
        """Get the list of sample queries for a data product.

        Args:
            dp_id (str): ID of the data product

        Returns:
            List[SampleQuery]: List of sample queries

        Raises:
            Exception: If the API request fails
        """
        response = requests.get(
            url=f'{self.protocol}://{self.host}/{self.DATA_PRODUCT_PATH}/{dp_id}/sampleQueries',
            auth=(self.username, self.password)
        )
        if not response.ok:
            raise Exception(f'Request returned code {response.status_code}.\nResponse body: {response.text}')
        return [SampleQuery.load(result) for result in response.json()]
    
    
    def get_materialized_view_refresh_metadata(self, dp_id: str, view_name: str) -> MaterializedViewRefreshMetadata:
        """Get refresh metadata for a materialized view.

        Args:
            dp_id (str): ID of the data product
            view_name (str): Name of the materialized view

        Returns:
            MaterializedViewRefreshMetadata: Metadata about the view's refresh status

        Raises:
            Exception: If the API request fails
        """
        response = requests.get(
            url=f'{self.protocol}://{self.host}/{self.DATA_PRODUCT_PATH}/{dp_id}/materializedViews/{view_name}/refreshMetadata',
            auth=(self.username, self.password)
        )
        if not response.ok:
            raise Exception(f'Request returned code {response.status_code}.\nResponse body: {response.text}')
        # response.json() will be None in scenario where no refresh has occurred yet
        if response.json() is None:
            return MaterializedViewRefreshMetadata(lastImport=None, incrementalColumn=None,refreshInterval=None,storageSchema=None,estimatedNextRefreshTime=None)
        return MaterializedViewRefreshMetadata.load(response.json())


    # --- domain API methods ---
    def create_domain(self, name: str, description: str=None, schema_location: str=None) -> Domain:
        """Create a new domain.

        Args:
            name (str): Name of the domain
            description (str, optional): Description of the domain. Defaults to None.
            schema_location (str, optional): Schema location for the domain. Defaults to None.

        Returns:
            Domain: The created domain

        Raises:
            Exception: If the API request fails
        """
        response = requests.post(
            url=f'{self.protocol}://{self.host}/{self.DOMAIN_PATH}',
            auth=(self.username, self.password),
            json={
                'name': name,
                'description': description,
                'schemaLocation': schema_location
            }
        )
        if not response.ok:
            raise Exception(f'Request returned code {response.status_code}.\nResponse body: {response.text}')
        return Domain.load(response.json())
    

    def delete_domain(self, domain_id: str):
        """Delete a domain.

        Args:
            domain_id (str): ID of the domain to delete

        Raises:
            Exception: If the API request fails
        """
        response = requests.delete(
            url=f'{self.protocol}://{self.host}/{self.DOMAIN_PATH}/{domain_id}',
            auth=(self.username, self.password)
        )
        if not response.ok:
            raise Exception(f'Request returned code {response.status_code}.\nResponse body: {response.text}')
    

    def update_domain(self, domain_id: str, description: str=None, schema_location: str=None) -> Domain:
        """Update a domain's properties.

        Args:
            domain_id (str): ID of the domain to update
            description (str, optional): New description. Defaults to None.
            schema_location (str, optional): New schema location. Defaults to None.

        Returns:
            Domain: The updated domain

        Raises:
            Exception: If the API request fails
        """
        response = requests.put(
            url=f'{self.protocol}://{self.host}/{self.DOMAIN_PATH}/{domain_id}',
            auth=(self.username, self.password),
            json={
                'description': description,
                'schemaLocation': schema_location
            }
        )
        if not response.ok:
            raise Exception(f'Request returned code {response.status_code}.\nResponse body: {response.text}')
        return Domain.load(response.json())
    

    def list_domains(self) -> List[Domain]:
        """Get a list of all domains.

        Returns:
            List[Domain]: List of all domains

        Raises:
            Exception: If the API request fails
        """
        response = requests.get(
            url=f'{self.protocol}://{self.host}/{self.DOMAIN_PATH}',
            auth=(self.username, self.password)
        )
        if not response.ok:
            raise Exception(f'Request returned code {response.status_code}.\nResponse body: {response.text}')
        return [Domain.load(result) for result in response.json()]


    def get_domain(self, domain_id: str) -> Domain:
        """Get details of a specific domain.

        Args:
            domain_id (str): ID of the domain to retrieve

        Returns:
            Domain: The requested domain

        Raises:
            Exception: If the API request fails
        """
        response = requests.get(
            url= f'{self.protocol}://{self.host}/{self.DOMAIN_PATH}/{domain_id}',
            auth=(self.username, self.password)
        )
        if not response.ok:
            raise Exception('bad request' + str(response))
        return Domain.load(response.json())


    # --- tags API methods ---
    def update_tags(self, dp_id: str, tag_values: List[str]) -> Tag:
        """Update tags for a data product.

        Args:
            dp_id (str): ID of the data product
            tag_values (List[str]): List of tag values to set

        Returns:
            Tag: The updated tags

        Raises:
            Exception: If the API request fails
        """
        response = requests.put(
            url=f'{self.protocol}://{self.host}/{self.DATA_PRODUCT_TAGS_PATH}/products/{dp_id}',
            auth=(self.username, self.password),
            json=[{"value": val} for val in tag_values]
        )
        if not response.ok:
            raise Exception(f'Request returned code {response.status_code}.\nResponse body: {response.text}')
        return [Tag.load(result) for result in response.json()]
        
        
    def get_tags(self, dp_id: str) -> List[Tag]:
        """Get tags for a data product.

        Args:
            dp_id (str): ID of the data product

        Returns:
            List[Tag]: List of tags

        Raises:
            Exception: If the API request fails
        """
        response = requests.get(
            url=f'{self.protocol}://{self.host}/{self.DATA_PRODUCT_TAGS_PATH}/products/{dp_id}',
            auth=(self.username, self.password)
        )
        if not response.ok:
            raise Exception(f'Request returned code {response.status_code}.\nResponse body: {response.text}')
        return [Tag.load(result) for result in response.json()]
    
    
    def delete_tag(self, tag_id: str, dp_id: str):
        """Delete a tag from a data product.

        Args:
            tag_id (str): ID of the tag to delete
            dp_id (str): ID of the data product

        Raises:
            Exception: If the API request fails
        """
        response = requests.delete(
            url=f'{self.protocol}://{self.host}/{self.DATA_PRODUCT_TAGS_PATH}/{tag_id}/products/{dp_id}',
            auth=(self.username, self.password)
        )
        if not response.ok:
            raise Exception(f'Request returned code {response.status_code}.\nResponse body: {response.text}')
    

    # --- workflow API methods ---
    def publish_data_product(self, dp_id: str, force: bool=False):
        """Publish a data product.

        Args:
            dp_id (str): ID of the data product to publish
            force (bool, optional): Whether to force publish. Defaults to False.

        Raises:
            Exception: If the API request fails
        """
        response = requests.post(
            url=f'{self.protocol}://{self.host}/{self.DATA_PRODUCT_PATH}/{dp_id}/workflows/publish',
            auth=(self.username, self.password),
            params={'force': force}
        )
        if not response.ok:
            raise Exception(f'Request returned code {response.status_code}.\nResponse body: {response.text}')
    

    def get_publish_data_product_status(self, dp_id: str) -> DataProductWorkflowStatus:
        """Get the status of a data product publish workflow.

        Args:
            dp_id (str): ID of the data product

        Returns:
            DataProductWorkflowStatus: Status of the publish workflow

        Raises:
            Exception: If the API request fails
        """
        response = requests.get(
            url=f'{self.protocol}://{self.host}/{self.DATA_PRODUCT_PATH}/{dp_id}/workflows/publish',
            auth=(self.username, self.password),
        )
        if not response.ok:
            raise Exception(f'Request returned code {response.status_code}.\nResponse body: {response.text}')
        
        return DataProductWorkflowStatus.load(response.json())
    

    def delete_data_product(self, dp_id: str, skip_objects_delete: bool=False):
        """Delete a data product.

        Args:
            dp_id (str): ID of the data product to delete
            skip_objects_delete (bool, optional): Whether to skip deleting Trino objects. Defaults to False.

        Raises:
            Exception: If the API request fails
        """
        response = requests.post(
            url=f'{self.protocol}://{self.host}/{self.DATA_PRODUCT_PATH}/{dp_id}/workflows/delete',
            auth=(self.username, self.password),
            params={'skipTrinoDelete': skip_objects_delete}
        )
        if not response.ok:
            raise Exception(f'Request returned code {response.status_code}.\nResponse body: {response.text}')
    

    def get_delete_data_product_status(self, dp_id: str) -> DataProductWorkflowStatus:
        """Get the status of a data product delete workflow.

        Args:
            dp_id (str): ID of the data product

        Returns:
            DataProductWorkflowStatus: Status of the delete workflow

        Raises:
            Exception: If the API request fails
        """
        response = requests.get(
            url=f'{self.protocol}://{self.host}/{self.DATA_PRODUCT_PATH}/{dp_id}/workflows/delete',
            auth=(self.username, self.password),
        )
        if not response.ok:
            raise Exception(f'Request returned code {response.status_code}.\nResponse body: {response.text}')
        
        return DataProductWorkflowStatus.load(response.json())
    


from dataclasses import dataclass
from typing import List
from starburst_data_products_client.shared.models import JsonDataClass


@dataclass
class DataProductWorkflowError(JsonDataClass):
    entityType: str
    entityName: str
    message: str


@dataclass
class DataProductWorkflowStatus(JsonDataClass):
    workflowType: str
    status: str
    errors: List[DataProductWorkflowError]
    isFinalStatus: bool



from dataclasses import dataclass
from starburst_data_products_client.sep.data import Column
from typing import List, Optional
from datetime import datetime
from starburst_data_products_client.shared.models import JsonDataClass


@dataclass
class View(JsonDataClass):
    name: str
    description: Optional[str]
    createdBy: str
    definitionQuery: str
    status: str
    columns: List[Column]
    markedForDeletion: bool
    createdAt: datetime
    updatedAt: datetime
    updatedBy: str
    publishedAt: Optional[datetime]
    publishedBy: Optional[str]
    matchesTrinoDefinition: Optional[bool]



from dataclasses import dataclass
from typing import Optional
from starburst_data_products_client.shared.models import PaginatedJsonDataClass

@dataclass
class Tag(PaginatedJsonDataClass):
    name: str
    color: str
    description: Optional[str]



from dataclasses import dataclass
from typing import List
from starburst_data_products_client.galaxy.models import Contact, Link
from datetime import datetime
from starburst_data_products_client.shared.models import PaginatedJsonDataClass

@dataclass
class DataProductCatalogDetails(PaginatedJsonDataClass):
    catalogId: str
    catalogName: str
    catalogKind: str
    localRegions: List[str]


@dataclass
class CreateDataProductResponse(PaginatedJsonDataClass):
    syncToken: str
    dataProductId: str
    name: str
    summary: str
    description: str
    catalog: DataProductCatalogDetails
    schemaName: str
    contacts: List[Contact]
    links: List[Link]
    defaultClusterId: str
    createdOn: datetime
    modifiedOn: datetime
    createdBy: Contact
    modifiedBy: Contact
